# Copyright (C) 2007-2010 Samuel Abels.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
Formatting logs into human readable reports.
"""

def _underline(text, line = '-'):
    return [text, line * len(text)]

def _get_action_name(action, retry = 0):
    if retry == 0:
        return action.get_name()
    return action.get_name() + ' (retry %d)' % retry

def _get_successful_logs_from_action(logger, action):
    return [l for l in logger.get_logs(action) if not l.has_error()]

def _get_error_logs_from_action(logger, action):
    return [l for l in logger.get_logs(action) if l.has_error()]

def status(logger):
    """
    Creates a one-line summary on the actions that were logged by the given
    Logger.

    @type  logger: Logger
    @param logger: The logger that recorded what happened in the queue.
    @rtype:  string
    @return: A string summarizing the status.
    """
    total     = len(logger.get_logged_actions())
    aborted   = len(logger.get_aborted_actions())
    succeeded = total - aborted
    if total == 0:
        return 'No actions done'
    elif total == 1 and succeeded == 1:
        return 'One action done (succeeded)'
    elif total == 1 and succeeded == 0:
        return 'One action done (failed)'
    elif total == succeeded:
        return '%d actions total (all succeeded)' % total
    elif succeeded == 0:
        return '%d actions total (all failed)' % total
    else:
        msg = '%d actions total (%d failed, %d succeeded)'
        return msg % (total, aborted, succeeded)

def summarize(logger):
    """
    Creates a short summary on the actions that were logged by the given
    Logger.

    @type  logger: Logger
    @param logger: The logger that recorded what happened in the queue.
    @rtype:  string
    @return: A string summarizing the status of every performed task.
    """
    summary = []
    for action in logger.get_logged_actions():
        for i, log in enumerate(logger.get_logs(action)):
            thestatus = log.has_error() and log.get_error(False) or 'ok'
            name      = log.get_name()
            if i > 0:
                name += ' (retry %d)' % i
            summary.append(name + ': ' + thestatus)
    return '\n'.join(summary)

def format(logger,
           show_successful = True,
           show_errors     = True,
           show_traceback  = True):
    """
    Prints a report of the actions that were logged by the given Logger.
    The report contains a list of successful actions, as well as the full
    error message on failed actions.

    @type  logger: Logger
    @param logger: The logger that recorded what happened in the queue.
    @rtype:  string
    @return: A string summarizing the status of every performed task.
    """
    output = []

    # Print failed actions.
    errors = logger.get_aborted_actions()
    if show_errors and errors:
        output += _underline('Failed actions:')
        for action in errors:
            for i, log in enumerate(logger.get_logs(action)):
                name = _get_action_name(action, i)
                if show_traceback:
                    output.append(name + ':')
                    output.append(log.get_error())
                else:
                    output.append(name + ': ' + log.get_error(False))
            output.append('')

    # Print successful actions.
    if show_successful:
        output += _underline('Successful actions:')
        for action in logger.get_successful_actions():
            n_errors = len(_get_error_logs_from_action(logger, action))
            if n_errors == 0:
                thestatus = ''
            elif n_errors == 1:
                thestatus = ' (required one retry)'
            else:
                thestatus = ' (required %d retries)' % n_errors
            output.append(_get_action_name(action) + thestatus)
        output.append('')

    return '\n'.join(output).strip()
